@@ -1,3 +1,4 @@
+from __future__ import annotations
 import io
 import os
 import subprocess
@@ -5,6 +6,12 @@ import ctypes
 import json
 import shlex
 import contextlib
+import pathlib
+import typing_extensions
+import toml
+import typing
+from pydantic.dataclasses import dataclass
+from pydantic import ValidationError
 from .defaults import get_julia_exe, get_project_args
 
 CF_TYPY_MODE = "TYPY_MODE"
@@ -16,6 +23,9 @@ CF_TYPY_JL_EXE = "TYPY_JL_EXE"
 CF_TYPY_JL_OPTS = "TYPY_JL_OPTS"
 CF_JNUMPY_HOME = "JNUMPY_HOME"
 
+# debug
+# os.environ[CF_TYPY_JL_OPTS] = "--compile=min -O0"
+
 julia_info_query = r"""
 import Libdl
 import Pkg
@@ -25,25 +35,55 @@ println(unsafe_string(Base.JLOptions().image_file))
 println(dirname(Pkg.project().path))
 """.replace('\n', ';').replace('\r', ';')
 
+TyPython_dir = os.path.abspath("./TyPython")
+
+class JuliaProjectDict(typing_extensions.TypedDict):
+    name: str
+    deps: typing.Dict[str, str]
+
+@dataclass
+class JuliaProject:
+    name: str
+    deps: typing.Dict[str, str]
+
+class ParseProjectResult:
+    proj: JuliaProject
+
+def check_project(fullproj: dict, result: ParseProjectResult) -> typing_extensions.TypeGuard[JuliaProjectDict]:
+    try:
+        proj: JuliaProject = JuliaProject.__pydantic_model__.parse_obj(fullproj) # type: ignore
+        result.proj = proj
+        return True
+    except ValidationError:
+        return False
+
 exec_template = r"""
-try
-    import TyPython.CPython
-    Base.@eval CPython.WITH_GIL() do
-        {}
-    end
-catch e
-    CPython.WITH_GIL() do
-        errmsg = CPython.Utils.capture_out() do
-            Base.showerror(stderr, e, catch_backtrace())
+begin
+    using TyPython.CPython
+    try
+        Base.@eval begin
+            __PY_GIL = CPython.GIL_BEGIN()
+            try
+                {}
+            finally
+                CPython.GIL_END(__PY_GIL)
+            end
+        end
+    catch e
+        CPython.WITH_GIL() do
+            errmsg = CPython.Utils.capture_out() do
+                Base.showerror(stderr, e, catch_backtrace())
+            end
+            sys = CPython.py_import("sys")
+            err_o = CPython.py_cast(CPython.Py, errmsg)
+            sys.stderr.write(err_o)
+            rethrow()
         end
-        sys = CPython.py_import("sys")
-        err_o = CPython.py_cast(CPython.Py, errmsg)
-        sys.stderr.write(err_o)
-        rethrow()
     end
 end
 """
 
+
 def args_from_config(exepath: str, args: list):
     args = [exepath] + args
 
@@ -64,7 +104,6 @@ def exec_julia(x):
 def include_src(src_file: str, file_path: str=""):
     """
     include julia module in src_file
-
     Arguments:
       src_file:
         the path of julia file releative to file path.
@@ -72,13 +111,65 @@ def include_src(src_file: str, file_path: str=""):
         should be `__file__`, empty in repl mode.
     """
     # activate project before include module
-    global project_dir
     file_dir = os.path.dirname(file_path)
     src_path = os.path.join(file_dir, src_file)
     src_path = os.path.abspath(src_path)
     activate_proj(file_dir)
     exec_julia("include({})".format(escape_string(src_path)))
-    activate_proj(project_dir)
+    activate_proj(default_project_dir)
+
+_path_to_modulenames: dict[str, tuple[str, str, str, JuliaProjectDict]] = {}
+
+def load_project(file_path: str="."):
+    """
+    include julia module in project
+
+    Arguments:
+      file_path(option):
+        should be `__file__`, empty in repl mode.
+    """
+    # activate project before include module
+    file_dir = pathlib.Path(file_path).absolute().parent.as_posix()
+    res = _path_to_modulenames.get(file_dir)
+    if res is not None:
+        return res
+
+    project_path = os.path.join(file_dir, "Project.toml")
+    fullproj = toml.load(str(project_path))
+    result = ParseProjectResult()
+    if not check_project(fullproj, result):
+        raise RuntimeError(f"{file_dir} does not have a Project.toml with a top-level"
+                           f"entry 'name = xxx' and the '[deps]' section.")
+    proj = result.proj
+    res = _path_to_modulenames[file_dir] = (proj.name, file_dir, project_path, fullproj)
+    with activate_project(project_path, file_dir, fullproj):
+        exec_julia("import {0}".format(fullproj["name"]))
+
+    return res
+
+@contextlib.contextmanager
+def activate_project(project_path: str, file_dir: str, fullproj: JuliaProjectDict):
+    if fullproj["deps"].pop("TyPython", None):
+        with open(project_path, 'w', encoding='utf-8') as f:
+                toml.dump(fullproj, f)
+    activate_proj(file_dir)
+    try:
+        fullproj['deps']['TyPython'] = "9c4566a2-237d-4c69-9a5e-9d27b7d0881b"
+        with open(project_path, 'w', encoding='utf-8') as f:
+            toml.dump(fullproj, f)
+        try:
+            yield
+        finally:
+            with open(project_path, 'w', encoding='utf-8') as f:
+                del fullproj['deps']["TyPython"]
+                toml.dump(fullproj, f)
+    finally:
+        activate_proj(default_project_dir)
+
+def init_project(file_path):
+    modulename, file_dir, project_path, fullproj = load_project(file_path)
+    with activate_project(project_path, file_dir, fullproj):
+        exec_julia("import {0};TyPython.CPython.init();{0}.init()".format(modulename))
 
 def activate_proj(proj_dir: str):
     global _activate_proj
@@ -97,7 +188,7 @@ def init_jl():
     global _eval_jl
     global _add_deps
     global _activate_proj
-    global project_dir
+    global default_project_dir
     if os.getenv(CF_TYPY_MODE) == CF_TYPY_MODE_JULIA:
         return
     elif os.getenv(CF_TYPY_MODE) == CF_TYPY_MODE_PYTHON:
@@ -115,7 +206,7 @@ def init_jl():
     jl_opts = shlex.split(os.getenv(CF_TYPY_JL_OPTS, ""))
     jl_opts_proj = get_project_args()
     cmd = [jl_exepath, jl_opts_proj, *jl_opts, '--startup-file=no', '-O0', '--compile=min', '-e', julia_info_query]
-    bindir, libpath, sysimage, project_dir = subprocess.run(
+    bindir, libpath, sysimage, default_project_dir = subprocess.run(
         cmd, check=True, capture_output=True, encoding='utf8'
     ).stdout.splitlines()
 
@@ -141,9 +232,9 @@ def init_jl():
         if not lib.jl_eval_string(rf"""
         try
             import Pkg
-            Pkg.activate({escape_string(project_dir)}, io=devnull)
+            Pkg.activate({escape_string(default_project_dir)}, io=devnull)
             if !haskey(Pkg.project().dependencies, "TyPython")
-                Pkg.add("TyPython", io=devnull)
+                Pkg.develop(path={escape_string(TyPython_dir)})
             end
             Pkg.instantiate()
             import TyPython
@@ -169,7 +260,7 @@ def init_jl():
                 Pkg.activate({escape_string(proj_dir)}, io=devnull)
                 Pkg.instantiate()
             catch err
-                showerror(stderr, err, catch_backtrace())
+                showerror(stdout, err, catch_backtrace())
                 rethrow()
             end
             """.encode('utf8')):
